/*
 * grayupload - implementation of the GNU Automatic FTP Uploads
 * Copyright (C) 2023 Sergey Poznyakoff
 *
 * Grayupload is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * Grayupload is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with grayupload. If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <grayupload.h>

const char *progname;

static struct optdef *
option_find (struct parseopt *po, char const *optstr, int single_dash)
{
  size_t optlen = strlen (optstr);
  int i;
  int found = 0;
  struct optdef *found_opt = NULL;

  for (i = 0; i < po->optcount; i++)
    {
      size_t namelen;
      struct optdef *opt = &po->optdef[po->optidx[i]];
      namelen = strlen (opt->opt_name);
      /* Ignore single-letter options when used with -- */
      if (namelen == 1 && !single_dash)
	continue;

      if (optlen <= namelen)
	{
	  if (memcmp (opt->opt_name, optstr, optlen) == 0)
	    {
	      switch (found)
		{
		case 0:
		  found_opt = opt;
		  found++;
		  if (optlen == namelen)
		    {
		      /* Exact match */
		      goto end;
		    }
		  break;

		case 1:
		  found++;
		  error ("option '-%*.*s' is ambiguous; possibilities:",
			 optlen, optlen, optstr);
		  fprintf (stderr, "-%s\n", found_opt->opt_name);
		  /* fall through */
		case 2:
		  fprintf (stderr, "-%s\n", opt->opt_name);
		}
	    }
	  else if (found)
	    break;
	}
    }
 end:
  switch (found)
    {
    case 0:
      error ("unrecognized option '-%*.*s'", optlen, optlen, optstr);
      break;

    case 1:
      while (found_opt > po->optdef && (found_opt->opt_flags & OPTFLAG_ALIAS))
	found_opt--;
      return found_opt;

    case 2:
      break;
    }
  return NULL;
}

static int header_col = 1;
static int opt_col = 2;
static int opt_doc_col = 29;
#define DEFAULT_RIGHT_MARGIN 79
static int rmargin = DEFAULT_RIGHT_MARGIN;

static void
detect_right_margin (void)
{
  struct winsize ws;
  unsigned r = 0;

  ws.ws_col = ws.ws_row = 0;
  if ((ioctl (1, TIOCGWINSZ, (char *) &ws) < 0) || ws.ws_col == 0)
    {
      char *p = getenv ("COLUMNS");
      if (p)
	{
	  unsigned long n;
	  char *ep;
	  errno = 0;
	  n = strtoul (p, &ep, 10);
	  if (!(errno || *ep || n > UINT_MAX))
	    r = n;
	}
      else
	r = DEFAULT_RIGHT_MARGIN;
    }
  else
    r = ws.ws_col;
  rmargin = r;
}

static void
format_para (int left, int right, char const *text, FILE *fp)
{
  int width = right - left - 1;

  if (strlen (text) < width)
    fprintf (fp, "%s\n", text);
  else
    {
      int len = 0;

      while (*text)
	{
	  size_t n = strcspn (text, " \t\n");

	  if (len + n > width)
	    {
	      fprintf (fp, "\n%*.*s", left, left, "");
	      len = 0;
	      while (*text && (*text == ' ' || *text == '\t'))
		;
	      continue;
	    }
	  else if (text[n] == '\n')
	    {
	      len = 0;
	      do
		{
		  n++;
		  len++;
		}
	      while (text[n] && (text[n] == ' ' || text[n] == '\t'));
	      len--;
	    }
	  else
	    len += n;

	  fwrite (text, n, 1, fp);
	  text += n;

	  while (*text && (*text == ' ' || *text == '\t'))
	    text++;

	  if (*text)
	    {
	      fputc (' ', fp);
	      len++;
	    }
	}
      fputc ('\n', fp);
    }
}

static void
parseopt_list_options (struct parseopt *po, FILE *fp)
{
  struct optdef *opt;
  int prev_doc = 0;

  for (opt = po->optdef; opt->opt_name; )
    {
      if (opt->opt_flags & OPTFLAG_DOC)
	{
	  if (!prev_doc)
	    fputc ('\n', fp);
	  fprintf (fp, "%*.*s", header_col, header_col, "");
	  format_para (header_col, rmargin, opt->opt_doc, fp);
	  fputc ('\n', fp);
	  opt++;
	  prev_doc = 1;
	}
      else
	{
	  struct optdef *cur_opt = opt;
	  int col;

	  col = fprintf (fp, "%*.*s-%s", opt_col, opt_col, "", cur_opt->opt_name);
	  for (opt++; opt->opt_name && (opt->opt_flags & OPTFLAG_ALIAS); opt++)
	    {
	      col += fprintf (fp, ", -%s", opt->opt_name);
	    }
	  if (cur_opt->opt_argdoc)
	    {
	      col += fprintf (fp, " %s", cur_opt->opt_argdoc);
	    }
	  if (col >= opt_doc_col)
	    {
	      fputc ('\n', fp);
	      col = 0;
	    }
	  for (; col < opt_doc_col; col++)
	    fputc (' ', fp);
	  format_para (opt_doc_col, rmargin, cur_opt->opt_doc, fp);
	  prev_doc = 0;
	}
    }
}

int
parseopt_next (struct parseopt *po, char **ret_arg)
{
  char *arg;
  struct optdef *opt;

  if (po->argi == po->argc)
    return EOF;

  arg = po->argv[po->argi++];
  if (po->eopt)
    {
      *ret_arg = arg;
      return 0;
    }

  if (arg[0] == '-')
    {
      int single_dash = 1;
      if (arg[1] == '-')
	{
	  if (arg[2] == 0)
	    {
	      if (po->argi == po->argc)
		return EOF;
	      po->eopt = 1;
	      *ret_arg = po->argv[po->argi++];
	      return 0;
	    }
	  else
	    {
	      arg++;
	      single_dash = 0;
	    }
	}

      opt = option_find (po, arg + 1, single_dash);
      if (!opt)
	{
	  exit (EX_USAGE);
	}
      if (opt->opt_argdoc)
	{
	  if (po->argi == po->argc)
	    {
	      error ("option '-%s' requires argument", opt->opt_name);
	      exit (EX_USAGE);
	    }
	  *ret_arg = po->argv[po->argi++];
	}
      return opt->opt_code;
    }
  *ret_arg = arg;
  return 0;
}

int
parseopt_init (struct parseopt *po, int argc, char **argv)
{
  int i, j, k;

  if ((progname = strrchr (argv[0], '/')) == NULL)
    progname = argv[0];
  else
    progname++;

  po->argc = argc;
  po->argv = argv;
  po->argi = 1;
  po->eopt = 0;

  /* Collect and sort actual options */
  po->optcount = 0;
  for (i = 0; po->optdef[i].opt_name; i++)
    if (!(po->optdef[i].opt_flags & OPTFLAG_DOC))
      po->optcount++;

  if (po->optcount == 0)
    return 0;

  po->optidx = xcalloc (po->optcount, sizeof (po->optidx[0]));
  for (i = 0; po->optdef[i].opt_flags & OPTFLAG_DOC; i++)
    ;

  j = 0;
  po->optidx[j++] = i;
  if (po->optcount > 1)
    {
      for (i++; po->optdef[i].opt_name; i++)
	{
	  if (po->optdef[i].opt_flags & OPTFLAG_DOC)
	    continue;
	  po->optidx[j] = i;
	  for (k = j; k > 0 &&
		 strcmp (po->optdef[po->optidx[k-1]].opt_name,
			 po->optdef[po->optidx[k]].opt_name) > 0; k--)
	    {
	      int tmp = po->optidx[k];
	      po->optidx[k] = po->optidx[k-1];
	      po->optidx[k-1] = tmp;
	    }
	  j++;
	}
    }
  return 0;
}

static char extra_help[] =
  "Long options can be preceded with -- instead of -. "
  "A -- alone marks end of options: the remaining arguments are treated as "
  "file names.";

void
usage (int err, struct parseopt *po)
{
  FILE *fp = err ? stderr : stdout;
  detect_right_margin ();
  fprintf (fp, "usage: %s %s\n", progname, po->argdoc);
  format_para (0, rmargin, po->descr, fp);
  parseopt_list_options (po, fp);
  fputc ('\n', fp);
  format_para (0, rmargin, extra_help, fp);
  fputc ('\n', fp);
  fprintf (fp, "Report bugs and suggestions to <%s>\n", PACKAGE_BUGREPORT);
#ifdef PACKAGE_URL
  fprintf (fp, "%s home page: <%s>\n", PACKAGE_NAME, PACKAGE_URL);
#endif
  exit (err);
}
